#Sensor fix
sensors_dict = {
    "SENTINEL2_L1C": {
        "sensor_biovar": ["Cab", "Cm", "Cw", "FVC", "LAI", "laiCab", "laiCm", "laiCw"],
        "scale_factor": 0.0001,
        "bandlist": [
            "B02",
            "B03",
            "B04",
            "B05",
            "B06",
            "B07",
            "B08",
            "B8A",
            "B11",
            "B12",
        ],
    },
    "SENTINEL2_L2A": {
        "sensor_biovar": [
            "Cab",
            "Cm",
            "Cw",
            "FVC",
            "LAI",
            "laiCab",
            "laiCm",
            "laiCw",
            "CNC_Cab",
            "CNC_Cprot",
            "mangrove_Cab",
            "mangrove_Cm",
            "mangrove_Cw",
            "mangrove_LAI",
        ],
        "scale_factor": 0.0001,
        "bandlist": [
            "B02",
            "B03",
            "B04",
            "B05",
            "B06",
            "B07",
            "B08",
            "B8A",
            "B11",
            "B12",
        ],
    },
    "SENTINEL3_OLCI_L1B": {
        "sensor_biovar": ["FAPAR", "FVC", "LAI", "LCC"],
        "scale_factor": 1,
        "bandlist": [
            "B01",
            "B02",
            "B03",
            "B04",
            "B05",
            "B06",
            "B07",
            "B08",
            "B09",
            "B10",
            "B11",
            "B12",
            "B13",
            "B14",
            "B15",
            "B16",
            "B17",
            "B18",
            "B19",
            "B20",
            "B21",
        ],
    },
    "SENTINEL3_SYN_L2_SYN": {
        "sensor_biovar": ["FAPAR", "FVC"],
        "scale_factor": 1,
        "bandlist": [
          "Syn_Oa01_reflectance",
          "Syn_Oa02_reflectance",
          "Syn_Oa03_reflectance",
          "Syn_Oa04_reflectance",
          "Syn_Oa05_reflectance",
          "Syn_Oa06_reflectance",
          "Syn_Oa07_reflectance",
          "Syn_Oa08_reflectance",
          "Syn_Oa09_reflectance",
          "Syn_Oa10_reflectance",
          "Syn_Oa11_reflectance",
          "Syn_Oa12_reflectance",
          "Syn_Oa16_reflectance",
          "Syn_Oa17_reflectance",
          "Syn_Oa18_reflectance",
          "Syn_Oa21_reflectance",
          "Syn_S1N_reflectance",
          "Syn_S2N_reflectance",
          "Syn_S3N_reflectance",
          "Syn_S5N_reflectance",
          "Syn_S6N_reflectance",
        ],
    },  
}



import openeo

test_udf = "Udf test passed"

udf_gpr = openeo.UDF(
    """

import importlib
import os
import sys
import time

dir_path = 'tmp/venv'
sys.path.insert(0, dir_path)

python_files = [f[:-3] for f in os.listdir(dir_path) if f.endswith('.py') and f !="__init__.py"]

for lib in python_files:
    globals()[lib] = __import__(lib)

from configparser import ConfigParser
from typing import Dict
from openeo.metadata import Band, CollectionMetadata
from openeo.udf import XarrayDataCube, inspect

import numpy as np
import xarray as xr
from pathlib import Path
from openeo.udf.debug import inspect

chunks = 128
def broadcaster(array,bands_from_model):
    return np.broadcast_to(array[:, np.newaxis, np.newaxis], (bands_from_model, chunks, chunks))

init_xr = xr.DataArray()
def apply_datacube(cube: xarray.DataArray, context: dict) -> xarray.DataArray:

    sensor = context["sensor"]
    variable = context["biovar"]
    model = globals()[sensor + "_" + variable]
    bands = len(model.mx_GREEN.ravel())

    hyp_ell_GREEN = broadcaster(model.hyp_ell_GREEN,bands)
    mx_GREEN = broadcaster(model.mx_GREEN.ravel(),bands)
    sx_GREEN = broadcaster(model.sx_GREEN.ravel(),bands)

    XDX_pre_calc_GREEN_broadcast = np.broadcast_to(model.XDX_pre_calc_GREEN.ravel()[:,np.newaxis,np.newaxis],(model.XDX_pre_calc_GREEN.shape[0],chunks,chunks))


    if sensor == "SENTINEL3_SYN_L2_SYN":
        pixel_spectra = (cube.values/10000)
    else:
        pixel_spectra = (cube.values)

    im_norm_ell2D_hypell  = ((pixel_spectra - mx_GREEN) / sx_GREEN) * hyp_ell_GREEN
    im_norm_ell2D  = ((pixel_spectra - mx_GREEN) / sx_GREEN)

    PtTPt = np.einsum('ijk,ijk->ijk', im_norm_ell2D_hypell, im_norm_ell2D) * (-0.5)
    PtTDX = np.einsum('ij,jkl->ikl', model.X_train_GREEN,im_norm_ell2D_hypell)
    arg1 = np.exp(PtTPt[0]) * model.hyp_sig_GREEN

    k_star_im = np.exp(PtTDX - (XDX_pre_calc_GREEN_broadcast * (0.5)))
    mean_pred = (np.einsum('ijk,i->jk',k_star_im, model.alpha_coefficients_GREEN.ravel()) * arg1) + model.mean_model_GREEN

    # Uncertainty calculation
    k_star_uncert_im = np.exp(PtTDX - (XDX_pre_calc_GREEN_broadcast * (0.5))) * arg1    
    k_star_uncert = np.expand_dims(k_star_uncert_im, axis=0)
    
    k_star_uncert_flat = k_star_uncert_im.reshape(514, -1) 
    Vvector = np.matmul(model.Linv_pre_calc_GREEN, k_star_uncert_flat)  
    
    V_norm2 = np.sum(Vvector*Vvector, axis=0)  
    
    diff = model.hyp_sig_GREEN - V_norm2
    Variance = np.sqrt(np.abs(diff)) 
    
    variance_band = Variance.reshape(128,128)
    stacked = np.stack([mean_pred, variance_band], axis=0)
    
    returned = xr.DataArray(
        stacked,
        dims=("band", "y", "x"), 
        coords={"band": ["mean", "variance"]}
    )
    return returned
""",
    context={"from_parameter": "context"},
)


udf_gpr_customized = """

import importlib
import os
import sys
import time

from configparser import ConfigParser
from typing import Dict
from openeo.metadata import Band, CollectionMetadata
from openeo.udf import XarrayDataCube, inspect

import numpy as np
import xarray as xr
from pathlib import Path
from openeo.udf.debug import inspect

bands = len(np.array({mx_GREEN_placeholder}).ravel())
chunks = 128

def broadcaster(array):
    array = np.array(array)
    return np.broadcast_to(array[:, np.newaxis, np.newaxis], (bands, chunks, chunks))

#hyperparameters that need broadcasting

hyp_ell_GREEN = broadcaster(np.array({hyp_ell_GREEN_placeholder}))
mx_GREEN = broadcaster(np.array({mx_GREEN_placeholder}).ravel())
sx_GREEN = broadcaster(np.array({sx_GREEN_placeholder}).ravel())
XDX_pre_calc_GREEN_broadcast = np.broadcast_to(np.array({XDX_pre_calc_GREEN_placeholder}).ravel()[:,np.newaxis,np.newaxis],
                                               (np.array({XDX_pre_calc_GREEN_placeholder}).shape[0],chunks,chunks))

#hyperparameters that don't need broadcasting
X_train_GREEN = np.array({X_train_GREEN_placeholder})
hyp_sig_GREEN = np.array({hyp_sig_GREEN_placeholder})
alpha_coefficients_GREEN = np.array({alpha_coefficients_GREEN_placeholder})
mean_model_GREEN = {mean_model_GREEN_placeholder}

init_xr = xr.DataArray()
def apply_datacube(cube: xarray.DataArray, context: dict) -> xarray.DataArray:

    pixel_spectra = (cube.values)

    im_norm_ell2D_hypell  = ((pixel_spectra - mx_GREEN) / sx_GREEN) * hyp_ell_GREEN
    im_norm_ell2D  = ((pixel_spectra - mx_GREEN) / sx_GREEN)
    
    inspect(data=[im_norm_ell2D_hypell.dtype], message="im_norm_ell2D_hypell.dtype")
    inspect(data=[im_norm_ell2D.dtype], message="im_norm_ell2D.dtype")
    inspect(data=[hyp_ell_GREEN.dtype], message="hyp_ell_GREEN.dtype")
    
    inspect(data=[im_norm_ell2D_hypell.shape], message="im_norm_ell2D_hypell.shape")
    inspect(data=[im_norm_ell2D.shape], message="im_norm_ell2D.shape")
    inspect(data=[hyp_ell_GREEN.shape], message="hyp_ell_GREEN.shape")
    
    PtTPt = np.einsum('ijk,ijk->ijk', im_norm_ell2D_hypell.astype(np.float16), im_norm_ell2D.astype(np.float16)) * (-0.5)
    inspect(data=[PtTPt.shape], message="PtTPt.shape")
    PtTDX = np.einsum('ij,jkl->ikl', X_train_GREEN.astype(np.float16),im_norm_ell2D_hypell.astype(np.float16))
    arg1 = np.exp(PtTPt[0]) * hyp_sig_GREEN

    k_star_im = np.exp(PtTDX - (XDX_pre_calc_GREEN_broadcast * (0.5)))
    mean_pred = (np.einsum('ijk,i->jk',k_star_im.astype(np.float16), alpha_coefficients_GREEN.ravel().astype(np.float16)) * arg1.astype(np.float16)) + mean_model_GREEN

    init_xr = np.clip(mean_pred, a_min=0, a_max=None)
    returned = xr.DataArray(init_xr)
    return returned
"""


def custom_model_import(user_module):
    custom_gpr = udf_gpr_customized.format(
        hyp_ell_GREEN_placeholder=repr(user_module.hyp_ell_GREEN.tolist()),
        mx_GREEN_placeholder=repr(user_module.mx_GREEN.tolist()),
        sx_GREEN_placeholder=repr(user_module.sx_GREEN.tolist()),
        XDX_pre_calc_GREEN_placeholder=repr(user_module.XDX_pre_calc_GREEN.tolist()),
        X_train_GREEN_placeholder=repr(user_module.X_train_GREEN.tolist()),
        hyp_sig_GREEN_placeholder=repr(user_module.hyp_sig_GREEN),
        alpha_coefficients_GREEN_placeholder=repr(
            user_module.alpha_coefficients_GREEN.tolist()
        ),
        mean_model_GREEN_placeholder=repr(user_module.mean_model_GREEN),
    )

    custom_udf = openeo.UDF(custom_gpr)

    return custom_udf


# %%



import openeo
import numpy as np
import scipy.signal
import importlib.util
from pyeogpr.sensors import sensors_dict
from pyeogpr.udfgpr import udf_gpr, custom_model_import
from pyeogpr.udfsgolay import udf_sgolay


class Datacube:
    """

    pyeogpr.Datacube
    ----------------

        sensor : SENTINEL2_L1C, SENTINEL2_L2A, SENTINEL3_OLCI_L1B, SENTINEL3_SYN_L2_SYN
            Satellite sensor to process the data with.

        biovar : Biophysical variable to process. The selected variable's map will be retrieved.
            Currently "built-in" variables available for each sensor:

            ======================  =====================================================
            Satellite Level         Available Products
            ======================  =====================================================
            SENTINEL2_L1C           Cab, Cm, Cw, FVC, LAI, laiCab, laiCm, laiCw
            SENTINEL2_L2A           Cab, Cm, Cw, FVC, LAI, laiCab, laiCm, laiCw, CNC_Cab, CNC_Cprot,
                                    mangrove_LAI, mangrove_Cm, mangrove_Cw, mangrove_Cab
            SENTINEL3_OLCI_L1B      FAPAR, FVC, LAI, LCC
            SENTINEL3_SYN_L2_SYN    FAPAR, FVC
            ======================  =====================================================

            - for own model, simply put the directory of your model

        bounding_box : list
            Your region of interest. Insert bbox as list. Can be selected from https://geojson.io/
            (e.g.: [-4.55, 42.73,-4.48, 42.77])

        temporal_extent : list
            Your temporal extent to be processed. (e.g.: ["2021-01-01", "2021-12-31"])

        cloudmask : Boolean
            If "True" the Sentinel 2 cloud mask will be applied (only to S2 data), with Gaussian convolution to have
            smoother edges when masking.


    """


    def __init__(self, sensor, biovar, bounding_box, temporal_extent, cloudmask=False, bands=None):
        self.connection = openeo.connect("https://openeo.dataspace.copernicus.eu").authenticate_oidc()
        self.sensor = sensor
        self.own_model = biovar if biovar.endswith(".py") else None
        self.biovar = "Own_variable" if self.own_model else biovar
        self.bounding_box = bounding_box
        self.temporal_extent = temporal_extent
        self.cloudmask = cloudmask
        self.sensors_dict = sensors_dict
        self.bands = bands or self.sensors_dict.get(sensor, {}).get("bandlist")
        self.scale_factor = None
        self.data = None
        self.masked_data = None
        self.gpr_cube = None
        self.gpr_cube_gapfilled = None
        self.models_url = "https://github.com/daviddkovacs/pyeogpr/raw/main/models/GPR_models_bulk.zip#tmp/venv"
        self.ex_memory = "2g"
        self.py_memory = "12g"
        self.spatial_extent = {
            "west": bounding_box[0],
            "south": bounding_box[1],
            "east": bounding_box[2],
            "north": bounding_box[3],
        }


    def _load_base_collection(self):
        scale = self.sensors_dict.get(self.sensor, {}).get("scale_factor", 0.0001)
        return self.connection.load_collection(
            self.sensor,
            spatial_extent=self.spatial_extent,
            temporal_extent=self.temporal_extent,
            bands=self.bands,
        ) * scale


    def _apply_s2_cloudmask(self, data):
        s2_cloudmask = self.connection.load_collection(
            "SENTINEL2_L2A", self.spatial_extent, self.temporal_extent, ["SCL"]
        )
        scl = s2_cloudmask.band("SCL")
        mask = ~((scl == 4) | (scl == 5))

        kernel = np.outer(
            scipy.signal.windows.gaussian(11, std=1.6),
            scipy.signal.windows.gaussian(11, std=1.6)
        )
        kernel /= kernel.sum()
        smoothed_mask = mask.apply_kernel(kernel) > 0.1

        return data.mask(smoothed_mask)


    def _load_custom_model(self):
        spec = importlib.util.spec_from_file_location("user_module", self.own_model)
        user_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(user_module)
        return custom_model_import(user_module)

    def _apply_gpr(self):
        if self.own_model:
            print("Applying user-defined model.")
            custom_udf = self._load_custom_model()
            return self.masked_data.apply_dimension(process=custom_udf, dimension="bands")
        else:
            print(f"Applying default GPR model: {self.biovar}")
            context = {"sensor": self.sensor, "biovar": self.biovar}
            return self.masked_data.apply_dimension(
                process=udf_gpr, dimension="bands", context=context
            ).filter_bands(bands=[self.bands[0],self.bands[1]])

    def construct_datacube(self, composite=None):
        """
        Build the datacube with optional temporal compositing and cloud masking.

        Parameters
        ----------
        composite : str, optional
            Temporal compositing interval (e.g., 'month', 'dekad').
        """
        data = self._load_base_collection()

        if self.cloudmask and "SENTINEL2" in self.sensor:
            data = self._apply_s2_cloudmask(data)
            print("Applied Sentinel-2 cloud mask.")
        else:
            print(f"Cloud masking not applied or not supported for {self.sensor}.")

        if composite:
            data = data.aggregate_temporal_period(composite, "mean")
            print(f"Applied temporal compositing: {composite} by mean.")

        self.data = data
        self.masked_data = data
        print("Datacube constructed.")


    def process_map(self, gapfill=False, fileformat="tiff"):
        """
        Process the datacube into maps, by applying GPR algorithm on the spectral image stack.

        Parameters
        ----------
        gapfill : bool, default=False
            Apply Savitzky-Golay interpolator for gap filling.
        fileformat : str, default='tiff'
            Output file format ('nc' or 'tiff').
        """
        gpr_cube = self._apply_gpr()

        if gapfill:
            print("Applying Savitzky-Golay gapfilling.")
            gpr_cube_gapfilled = gpr_cube.apply_dimension(process=udf_sgolay, dimension="t")
            outputfile = f"{self.sensor}_{self.biovar}_GF.{fileformat}"
            title = f"{self.sensor} {self.biovar} gapfilled"
            cube_to_execute = gpr_cube_gapfilled
        else:
            outputfile = f"{self.sensor}_{self.biovar}.{fileformat}" if not self.own_model else "user_defined_product.{fileformat}"
            title = f"{self.sensor}_{self.biovar}" if not self.own_model else "User defined product"
            cube_to_execute = gpr_cube

        job_options = {"executor-memory": self.ex_memory,
                       "python-memory": self.py_memory}
        if not self.own_model:
            job_options["udf-dependency-archives"] = [self.models_url]

        cube_to_execute.execute_batch(
            title=title,
            outputfile=outputfile,
            job_options=job_options,
        )

        self.gpr_cube = gpr_cube
        if gapfill:
            self.gpr_cube_gapfilled = gpr_cube_gapfilled
